@@ -22,7 +22,7 @@ from mpcrl.core.experience import ExperienceReplay
 from mpcrl.core.exploration import ExplorationStrategy
 from mpcrl.core.parameters import LearnableParametersDict
 from mpcrl.core.update import UpdateStrategy
-from mpcrl.util.random import generate_seeds
+from mpcrl.util.iters import generate_seeds
 
 ExpType = TypeVar("ExpType")
 
@@ -94,13 +94,7 @@ class LearningAgent(
             Name of the agent. If `None`, one is automatically created from a counter of
             the class' instancies.
         """
-        Agent.__init__(
-            self,
-            mpc=mpc,
-            fixed_parameters=fixed_parameters,
-            warmstart=warmstart,
-            name=name,
-        )
+        Agent.__init__(self, mpc, fixed_parameters, warmstart, name)
         LearningAgentCallbacks.__init__(self)
         self._raises: bool = True
         self._learnable_pars = learnable_parameters
@@ -110,7 +104,7 @@ class LearningAgent(
         if exploration is not None:
             self._exploration = exploration
         if not isinstance(update_strategy, UpdateStrategy):
-            update_strategy = UpdateStrategy(update_strategy)
+            update_strategy = UpdateStrategy(update_strategy, "on_timestep_end")
         self._update_strategy = update_strategy
         self.establish_callback_hooks()
 
@@ -182,21 +176,15 @@ class LearningAgent(
         UpdateError or UpdateWarning
             Raises the error or the warning (depending on `raises`) if the update fails.
         """
-        # prepare for training start
         self._raises = raises
-        returns = np.zeros(episodes, dtype=float)
+        returns = np.zeros(episodes, float)
         self.on_training_start(env)
 
         for episode, current_seed in zip(range(episodes), generate_seeds(seed)):
-            self.reset(seed=current_seed)
+            self.reset(current_seed)
             state, _ = env.reset(seed=current_seed, options=env_reset_options)
             self.on_episode_start(env, episode)
-            returns[episode] = self.train_one_episode(
-                env=env,
-                episode=episode,
-                init_state=state,
-                raises=raises,
-            )
+            returns[episode] = self.train_one_episode(env, episode, state, raises)
             self.on_episode_end(env, episode, returns[episode])
 
         self.on_training_end(env, returns)
@@ -263,8 +251,8 @@ class LearningAgent(
         # hook updates (always necessary)
         assert self._update_strategy.hook in {
             "on_episode_end",
-            "on_env_step",
-        }, "Updates can be hooked only to episode_end or env_step."
+            "on_timestep_end",
+        }, "Updates can be hooked only to episode_end or on_timestep_end."
         args_idx, kwargs_keys = (
             (1, ("episode",))
             if self._update_strategy.hook == "on_episode_end"
