@@ -16,7 +16,6 @@ import casadi as cs
 import numba as nb
 import numpy as np
 import numpy.typing as npt
-from csnlp.util.math import prod
 from csnlp.wrappers import Mpc, NlpSensitivity
 from gymnasium import Env
 from scipy.linalg import cho_solve, lstsq
@@ -30,7 +29,7 @@ from mpcrl.core.learning_rate import LearningRate
 from mpcrl.core.parameters import LearnableParametersDict
 from mpcrl.core.schedulers import Scheduler
 from mpcrl.core.update import UpdateStrategy
-from mpcrl.util.math import cholesky_added_multiple_identities, monomial_powers
+from mpcrl.util.math import cholesky_added_multiple_identities, monomials_basis_function
 
 ExpType: TypeAlias = Tuple[
     npt.NDArray[np.floating],  # rollout's costs
@@ -40,32 +39,6 @@ ExpType: TypeAlias = Tuple[
 ]
 
 
-@nb.njit
-def _consolidate_rollout(
-    rollout: List[Tuple[ObsType, ActType, float, ObsType, npt.NDArray[np.floating]]],
-    ns: int,
-    na: int,
-) -> Tuple[
-    int,
-    npt.NDArray[np.floating],
-    npt.NDArray[np.floating],
-    npt.NDArray[np.floating],
-    npt.NDArray[np.floating],
-]:
-    """Internal utility to convert a rollout list to arrays."""
-    N = len(rollout)
-    S = np.empty((N, ns))
-    E = np.empty((N, na, 1))  # additional dim required for Psi
-    L = np.empty(N)
-    sol_vals = np.empty((N, rollout[0][-1].size))
-    for i, (s, e, cost, _, sol_val) in enumerate(rollout):
-        S[i] = s.reshape(-1)  # type: ignore[union-attr]
-        E[i] = e
-        L[i] = cost
-        sol_vals[i] = sol_val.reshape(-1)
-    return N, S, E, L, sol_vals
-
-
 class LstdDpgAgent(RlLearningAgent[SymType, ExpType, LrType], Generic[SymType, LrType]):
     """Least-Squares Temporal Difference (LSTD) Deterministic Policy Gradient (DPG)
     agent, as first introduced in [1] as its stochastic counterpart, and then refined in
@@ -90,7 +63,7 @@ class LstdDpgAgent(RlLearningAgent[SymType, ExpType, LrType], Generic[SymType, L
     """
 
     __slots__ = (
-        "_dpidtheta",
+        "_sensitivity",
         "_Phi",
         "rollout_length",
         "_rollout",
@@ -205,7 +178,7 @@ class LstdDpgAgent(RlLearningAgent[SymType, ExpType, LrType], Generic[SymType, L
             where `s` is the state, `v` are the weights, and `Phi(s)` is the state
             feature vector. This function is assumed to have one input and one output.
             By default, if not provided, it is designed as all monomials of the state
-            with degrees <= 2 (see `LstdDpgAgent.monomials_state_features`).
+            with degrees <= 2 (see `mpcrl.util.math.monomials_basis_function`).
         lstsq_cond : float, optional
             Conditional number to be passed to `scipy.linalg.lstsq`. By default, `None`.
         linsolver : "csparse" or "qr" or "mldivide", optional
@@ -232,22 +205,22 @@ class LstdDpgAgent(RlLearningAgent[SymType, ExpType, LrType], Generic[SymType, L
         if not isinstance(update_strategy, UpdateStrategy):
             update_strategy = UpdateStrategy(update_strategy, "on_episode_end")
         super().__init__(
-            mpc=mpc,
-            update_strategy=update_strategy,
-            discount_factor=discount_factor,
-            learning_rate=learning_rate,  # type: ignore[arg-type]
-            learnable_parameters=learnable_parameters,
-            fixed_parameters=fixed_parameters,
-            exploration=exploration,
-            experience=experience,
-            max_percentage_update=max_percentage_update,
-            warmstart=warmstart,
-            name=name,
+            mpc,
+            update_strategy,
+            discount_factor,
+            learning_rate,  # type: ignore[arg-type]
+            learnable_parameters,
+            fixed_parameters,
+            exploration,
+            experience,
+            max_percentage_update,
+            warmstart,
+            name,
         )
         # initialize derivatives, state feature vector and hessian approximation
-        self._dpidtheta = self._init_dpg_derivatives(linsolver)
+        self._sensitivity = self._init_sensitivities(linsolver)
         self._Phi = (
-            LstdDpgAgent.monomials_state_features(mpc.ns, mpc.sym_type.__name__, 0, 2)
+            monomials_basis_function(mpc.ns, 0, 2)
             if state_features is None
             else state_features
         )
@@ -258,7 +231,7 @@ class LstdDpgAgent(RlLearningAgent[SymType, ExpType, LrType], Generic[SymType, L
         self.cho_solve_kwargs = cho_solve_kwargs
         # initialize others
         self.lstsq_cond = lstsq_cond
-        self.rollout_length = rollout_length
+        self.rollout_length = rollout_length or float("+inf")
         self._rollout: List[
             Tuple[
                 ObsType,
@@ -278,7 +251,7 @@ class LstdDpgAgent(RlLearningAgent[SymType, ExpType, LrType], Generic[SymType, L
     def update(self) -> Optional[str]:
         sample = self.experience.sample()
 
-        # congegrate sample's rollout into a unique least-squares problem. Since the
+        # congregate sample's rollout into a unique least-squares problem. Since the
         # sample is usually shortish, no point in using nb.jit
         L_, Phi_, Psi_, dpidtheta_ = [], [], [], []
         mask_phi = [False]
@@ -297,16 +270,16 @@ class LstdDpgAgent(RlLearningAgent[SymType, ExpType, LrType], Generic[SymType, L
         Phi_diff = self.discount_factor * Phi[mask_phi[:-1]] - Phi[mask_phi[1:]]
         A_v = Phi[mask_phi[1:]].T @ -Phi_diff
         b_v = Phi[mask_phi[1:]].T @ L
-        v = lstsq(
-            A_v, b_v, cond=self.lstsq_cond, lapack_driver="gelsy", check_finite=False
-        )[0]
+        v = lstsq(A_v, b_v, self.lstsq_cond, lapack_driver="gelsy", check_finite=False)[
+            0
+        ]
 
         # compute CAFA weights w
         A_w = Psi.T @ Psi
         b_w = Psi.T @ (L + Phi_diff @ v)
-        w = lstsq(
-            A_w, b_w, cond=self.lstsq_cond, lapack_driver="gelsy", check_finite=False
-        )[0]
+        w = lstsq(A_w, b_w, self.lstsq_cond, lapack_driver="gelsy", check_finite=False)[
+            0
+        ]
 
         # compute policy gradient
         dJdtheta = (dpidtheta @ dpidtheta.transpose((0, 2, 1))).sum(0) @ w
@@ -336,11 +309,12 @@ class LstdDpgAgent(RlLearningAgent[SymType, ExpType, LrType], Generic[SymType, L
 
         while not (truncated or terminated):
             # compute V(s) (perturbed and not perturbed)
-            action, sol = self.state_value(state, deterministic=False)
-            action_opt, sol_opt = self.state_value(state, deterministic=True)
+            action, sol = self.state_value(state, False)
+            action_opt, sol_opt = self.state_value(state, True)
 
             # step the system with the action just computed
             state_new, cost, truncated, terminated, _ = env.step(action)
+            self.on_env_step(env, episode, timestep)
 
             # store transition in experience
             if sol.success and sol_opt.success:
@@ -361,78 +335,36 @@ class LstdDpgAgent(RlLearningAgent[SymType, ExpType, LrType], Generic[SymType, L
             timestep += 1
 
             # first, check if current rollout has reached its length, and only then
-            # invoke on_env_step (as it might trigger an update)
-            if (
-                self.rollout_length is not None
-                and len(self._rollout) >= self.rollout_length
-            ):
+            # invoke on_timestep_end (as it might trigger an update)
+            if len(self._rollout) >= self.rollout_length:
                 self._consolidate_rollout_into_memory()
-            self.on_env_step(env, episode, timestep)
+            self.on_timestep_end(env, episode, timestep)
 
         # consolidate rollout at the end of episode, if no length was specified
         if self.rollout_length is None:
             self._consolidate_rollout_into_memory()
         return rewards
 
-    @staticmethod
-    def monomials_state_features(
-        ns: int,
-        sym_type: Literal["SX", "MX"],
-        mindegree: int,
-        maxdegree: int,
-    ) -> cs.Function:
-        """Gets the state feature vector composed of all monomials of degree in the
-        given range.
-
-        Parameters
-        ----------
-        ns : int
-            Number of states.
-        sym_type : {"MX" or "SX"}
-            CasADi symbolic type to be used to build the function.
-        mindegree : int
-            Minimum degree of monomials (included).
-        maxdegree : int
-            Maximum degree of monomials (included).
-
-        Returns
-        -------
-        casadi.Function
-            A casadi function of the form `Phi(s)`, where `s` is the state and `Phi` the
-            state feature vector.
-        """
-        s: SymType = getattr(cs, sym_type).sym("x", ns, 1)
-        y = cs.vertcat(
-            *(
-                prod(s**p)
-                for k in range(mindegree, maxdegree + 1)
-                for p in monomial_powers(ns, k)
-            )
-        )
-        return cs.Function("Phi", [s], [y], ["s"], ["Phi(s)"])
-
-    def _init_dpg_derivatives(self, linsolvertype: str) -> cs.Function:
+    def _init_sensitivities(self, linsolvertype: str) -> cs.Function:
         """Internal utility to compute the derivatives w.r.t. the learnable parameters
         and other functions in order to estimate the policy gradient."""
         nlp = self._V.nlp
-        y = nlp.primal_dual_vars()
-        theta = cs.vertcat(*self._learnable_pars.sym.values())
-        u0 = cs.vertcat(*self._V.first_actions.values())
-        all_syms = cs.vertcat(
-            nlp.p, nlp.x, nlp.lam_g, nlp.lam_h, nlp.lam_lbx, nlp.lam_ubx
-        )
+        y = nlp.primal_dual
+        theta = cs.vcat(self._learnable_pars.sym.values())
+        u0 = cs.vcat(self._V.first_actions.values())
+        x_lam_p = cs.vertcat(nlp.primal_dual, nlp.p)
 
         # compute first bunch of derivatives
-        nlp_ = NlpSensitivity(nlp, target_parameters=theta)
+        nlp_ = NlpSensitivity(nlp, theta)
         Kt = nlp_.jacobians["K-p"].T
         Ky = nlp_.jacobians["K-y"].T
         dydu0 = cs.evalf(cs.jacobian(u0, y)).T
 
         # convert SX to MX (so that we can use the linsolver)
         if nlp.sym_type is cs.SX:
-            all_syms, all_syms_sx = cs.MX.sym("in", *all_syms.shape), all_syms
-            Kt = cs.Function("dKdtheta", [all_syms_sx], [Kt]).wrap()(all_syms)
-            Ky = cs.Function("dKdy", [all_syms_sx], [Ky]).wrap()(all_syms)
+            x_lam_p, x_lam_p_sx = cs.MX.sym("in", *x_lam_p.shape), x_lam_p
+            Kt = cs.Function("dKdtheta", [x_lam_p_sx], [Kt]).wrap()(x_lam_p)
+            Ky = cs.Function("dKdy", [x_lam_p_sx], [Ky]).wrap()(x_lam_p)
 
         # compute last bunch derivative and convert to function for faster runtime
         if linsolvertype == "mldivide":
@@ -441,11 +373,24 @@ class LstdDpgAgent(RlLearningAgent[SymType, ExpType, LrType], Generic[SymType, L
             linsolver = cs.Linsol("linsolver", linsolvertype, Ky.sparsity())
             q = linsolver.solve(Ky, dydu0)
         dpidtheta = -Kt @ q
-        return cs.Function("dpidtheta", [all_syms], [dpidtheta])
+        sensitivity = cs.Function("dpidtheta", (x_lam_p,), (dpidtheta,))
+
+        # wrap to conveniently return arrays. Casadi does not support tensors with more
+        # than 2 dims, so dpidtheta gets squished in the 3rd dim and needs reshaping
+        ntheta, na = dpidtheta.shape
+
+        def func(sol_values: cs.DM, N: int) -> np.ndarray:
+            return (
+                sensitivity(sol_values.T)
+                .full()
+                .reshape(ntheta, na, N, order="F")
+                .transpose((2, 0, 1))
+            )
+
+        return func
 
     def _consolidate_rollout_into_memory(self) -> None:
-        """Internal utility to compact the current rollout into a single item in
-        memory."""
+        """Internal utility to compact current rollout into a single item in memory."""
         # convert to arrays
         N, S, E, L, vals = _consolidate_rollout(self._rollout, self._V.ns, self._V.na)
 
@@ -455,17 +400,31 @@ class LstdDpgAgent(RlLearningAgent[SymType, ExpType, LrType], Generic[SymType, L
 
         # compute dpidtheta and Psi (casadi does not support tensors with more than 2
         # dims, so dpidtheta gets squished in the third dim and needs to be reshaped)
-        ntheta, na = self._dpidtheta.size_out(0)
-        dpidtheta = (
-            self._dpidtheta(vals.T)
-            .full()
-            .reshape(ntheta, na, N, order="F")
-            .transpose((2, 0, 1))
-        )
-        Psi = (dpidtheta @ E).reshape(N, ntheta)
+        dpidtheta = self._sensitivity(vals, N)
+        Psi = (dpidtheta @ E).reshape(N, dpidtheta.shape[1])
 
         # save to memory and clear rollout
-        super().store_experience((L, Phi, Psi, dpidtheta))
+        self.store_experience((L, Phi, Psi, dpidtheta))
         self._rollout.clear()
         if self.policy_performances is not None:
             self.policy_performances.append(L.sum())
+
+
+@nb.njit
+def _consolidate_rollout(
+    rollout: List[Tuple[ObsType, ActType, float, ObsType, np.ndarray]],
+    ns: int,
+    na: int,
+) -> Tuple[int, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
+    """Internal utility to convert a rollout list to arrays."""
+    N = len(rollout)
+    S = np.empty((N, ns))
+    E = np.empty((N, na, 1))  # additional dim required for Psi
+    L = np.empty(N)
+    sol_vals = np.empty((N, rollout[0][-1].size))
+    for i, (s, e, cost, _, sol_val) in enumerate(rollout):
+        S[i] = s.reshape(-1)  # type: ignore[union-attr]
+        E[i] = e
+        L[i] = cost
+        sol_vals[i] = sol_val.reshape(-1)
+    return N, S, E, L, sol_vals
