@@ -56,13 +56,7 @@ class LstdQLearningAgent(
         132-137. IEEE.
     """
 
-    __slots__ = (
-        "_dQdtheta",
-        "_d2Qdtheta2",
-        "td_errors",
-        "cho_maxiter",
-        "cho_solve_kwargs",
-    )
+    __slots__ = ("_sensitivities", "td_errors", "cho_maxiter", "cho_solve_kwargs")
 
     def __init__(
         self,
@@ -156,52 +150,28 @@ class LstdQLearningAgent(
             the class' instancies.
         """
         super().__init__(
-            mpc=mpc,
-            update_strategy=update_strategy,
-            discount_factor=discount_factor,
-            learning_rate=learning_rate,  # type: ignore[arg-type]
-            learnable_parameters=learnable_parameters,
-            fixed_parameters=fixed_parameters,
-            exploration=exploration,
-            experience=experience,
-            max_percentage_update=max_percentage_update,
-            warmstart=warmstart,
-            name=name,
+            mpc,
+            update_strategy,
+            discount_factor,
+            learning_rate,  # type: ignore[arg-type]
+            learnable_parameters,
+            fixed_parameters,
+            exploration,
+            experience,
+            max_percentage_update,
+            warmstart,
+            name,
         )
-        self._dQdtheta, self._d2Qdtheta2 = self._init_Q_derivatives(hessian_type)
+        self._sensitivities = self._init_sensitivities(hessian_type)
         self.cho_maxiter = cho_maxiter
         if cho_solve_kwargs is None:
             cho_solve_kwargs = {"check_finite": False}
         self.cho_solve_kwargs = cho_solve_kwargs
         self.td_errors: Optional[List[float]] = [] if record_td_errors else None
 
-    def store_experience(  # type: ignore[override]
-        self, cost: SupportsFloat, solQ: Solution[SymType], solV: Solution[SymType]
-    ) -> None:
-        """Stores the gradient and hessian for the current transition in memoru.
-
-        Parameters
-        ----------
-        cost : float
-            The cost of this state transition.
-        solQ : Solution[SymType]
-            The solution to `Q(s,a)`.
-        solV : Solution[SymType]
-            The solution to `V(s+)`.
-        """
-        sol_values = solQ.all_vals
-        dQ: npt.NDArray[np.floating] = self._dQdtheta(sol_values).full().reshape(-1, 1)
-        ddQ: npt.NDArray[np.floating] = self._d2Qdtheta2(sol_values).full()
-        td_error: float = cost + self.discount_factor * solV.f - solQ.f
-        g = -td_error * dQ
-        H = dQ @ dQ.T - td_error * ddQ
-        if self.td_errors is not None:
-            self.td_errors.append(td_error)
-        return super().store_experience((g, H))
-
     def update(self) -> Optional[str]:
         sample = self.experience.sample()
-        gradient, Hessian = (np.mean(tuple(o), axis=0) for o in zip(*sample))
+        gradient, Hessian = (np.mean(tuple(o), 0) for o in zip(*sample))
         R = cholesky_added_multiple_identities(Hessian, maxiter=self.cho_maxiter)
         step = cho_solve((R, True), gradient, **self.cho_solve_kwargs).reshape(-1)
         return self._do_gradient_update(step)
@@ -219,7 +189,7 @@ class LstdQLearningAgent(
         state = init_state
 
         # solve for the first action
-        action, solV = self.state_value(state, deterministic=False)
+        action, solV = self.state_value(state, False)
         if not solV.success:
             self.on_mpc_failure(episode, None, solV.status, raises)
 
@@ -228,44 +198,69 @@ class LstdQLearningAgent(
             solQ = self.action_value(state, action)
 
             # step the system with action computed at the previous iteration
-            state, cost, truncated, terminated, _ = env.step(action)
+            new_state, cost, truncated, terminated, _ = env.step(action)
+            self.on_env_step(env, episode, timestep)
 
-            # compute V(s+)
-            action, solV = self.state_value(state, deterministic=False)
-            if solQ.success and solV.success:
-                self.store_experience(cost, solQ, solV)
-            else:
-                self.on_mpc_failure(episode, timestep, solV.status, raises)
+            # compute V(s+) and store transition
+            new_action, solV = self.state_value(new_state, False)
+            if not self._try_store_experience(cost, solQ, solV):
+                self.on_mpc_failure(
+                    episode, timestep, f"{solQ.status}/{solV.status}", raises
+                )
 
             # increase counters
+            state = new_state
+            action = new_action
             rewards += float(cost)
             timestep += 1
-            self.on_env_step(env, episode, timestep)  # better to call this at the end
+            self.on_timestep_end(env, episode, timestep)
         return rewards
 
-    def _init_Q_derivatives(
-        self, hessian_type: Literal["approx", "full"]
-    ) -> Tuple[cs.Function, cs.Function]:
+    def _init_sensitivities(self, hess_type: Literal["approx", "full"]) -> cs.Function:
         """Internal utility to compute the derivative of Q(s,a) w.r.t. the learnable
         parameters, a.k.a., theta."""
-        theta = cs.vertcat(*self._learnable_pars.sym.values())
+        theta = cs.vcat(self._learnable_pars.sym.values())
         nlp = self._Q.nlp
-        nlp_ = NlpSensitivity(nlp, target_parameters=theta)
+        nlp_ = NlpSensitivity(nlp, theta)
         Lt = nlp_.jacobians["L-p"]  # a.k.a., dQdtheta
         Ltt = nlp_.hessians["L-pp"]  # a.k.a., approximated d2Qdtheta2
-        if hessian_type == "approx":
+        if hess_type == "approx":
             d2Qdtheta2 = Ltt
-        elif hessian_type == "full":
+        elif hess_type == "full":
             dydtheta, _ = nlp_.parametric_sensitivity(second_order=False)
             d2Qdtheta2 = dydtheta.T @ nlp_.jacobians["K-p"] + Ltt
         else:
-            raise ValueError(f"Invalid type of hessian; got {hessian_type}.")
+            raise ValueError(f"Invalid type of hessian; got {hess_type}.")
 
-        # convert to functions (much faster runtime)
-        input = cs.vertcat(nlp.p, nlp.x, nlp.lam_g, nlp.lam_h, nlp.lam_lbx, nlp.lam_ubx)
-        dQdtheta_ = cs.Function("dQdtheta", [input], [Lt])
-        d2Qdtheta2_ = cs.Function("d2Qdtheta2", [input], [d2Qdtheta2])
-        assert (
-            not dQdtheta_.has_free() and not d2Qdtheta2_.has_free()
-        ), "Internal error in Q derivatives."
-        return dQdtheta_, d2Qdtheta2_
+        # convert to function (much faster runtime)
+        x_lam_p = cs.vertcat(nlp.primal_dual, nlp.p)
+        sensitivities = cs.Function("dQ", (x_lam_p,), (Lt, d2Qdtheta2))
+        assert not sensitivities.has_free(), "Internal error in Q sensitivities."
+
+        # wrap to conveniently return numpy arrays
+        def func(sol_values: cs.DM) -> Tuple[np.ndarray, np.ndarray]:
+            dQ, ddQ = sensitivities(sol_values)
+            return dQ.full().reshape(-1, 1), ddQ.full()
+
+        return func
+
+    def _try_store_experience(
+        self, cost: SupportsFloat, solQ: Solution[SymType], solV: Solution[SymType]
+    ) -> bool:
+        """Internal utility that tries to store the gradient and hessian for the current
+        transition in memory, if both V and Q were successful; otherwise, does not store
+        it. Returns whether it was successful or not."""
+        if solQ.success and solV.success:
+            sol_values = solQ.all_vals
+            dQ, ddQ = self._sensitivities(sol_values)
+            td_error = cost + self.discount_factor * solV.f - solQ.f
+            g = -td_error * dQ
+            H = dQ @ dQ.T - td_error * ddQ
+            self.store_experience((g, H))
+            success = True
+        else:
+            td_error = np.nan
+            success = False
+        if self.td_errors is not None:
+            self.td_errors.append(td_error)
+        return success
