@@ -1,7 +1,7 @@
 from itertools import chain
 from typing import Deque, Iterable, Iterator, Optional, TypeVar, Union
 
-from mpcrl.util.random import np_random
+from csnlp.util.random import np_random
 
 ExpType = TypeVar("ExpType")
 
@@ -21,8 +21,7 @@ class ExperienceReplay(Deque[ExpType]):
         include_last: Union[int, float] = 0,
         seed: Optional[int] = None,
     ) -> None:
-        """
-        Instantiate the container for experience replay memory.
+        """Instantiate the container for experience replay memory.
 
         Parameters
         ----------
@@ -50,8 +49,7 @@ class ExperienceReplay(Deque[ExpType]):
         self.np_random = np_random(seed)
 
     def sample(self) -> Iterator[ExpType]:
-        """
-        Samples the experience memory and yields the sampled items.
+        """Samples the experience memory and yields the sampled items.
 
         Returns
         -------
@@ -80,5 +78,5 @@ class ExperienceReplay(Deque[ExpType]):
 
         # get last n indices and the sampled indices from the remaining
         last = range(L - last_n, L)
-        sampled = self.np_random.choice(range(L - last_n), n - last_n, replace=False)
+        sampled = self.np_random.choice(range(L - last_n), n - last_n, False)
         yield from (self[i] for i in chain(last, sampled))
