@@ -1,8 +1,10 @@
 from itertools import combinations
 from typing import Optional, Tuple, Union
 
+import casadi as cs
 import numpy as np
 import numpy.typing as npt
+from csnlp.util.math import prod
 from scipy.linalg import solve_discrete_are
 from scipy.special import comb
 
@@ -148,7 +150,7 @@ def nchoosek(n: Union[int, npt.ArrayLike], k: int) -> Union[int, np.ndarray]:
         combinations or the combinations in a matrix.
     """
     return (
-        comb(n, k, exact=True)
+        comb(n, k, True)
         if isinstance(n, int)
         else np.row_stack(list(combinations(np.asarray(n).reshape(-1), k)))
     )
@@ -173,11 +175,40 @@ def monomial_powers(d: int, k: int) -> npt.NDArray[np.int64]:
     m = nchoosek(k + d - 1, d - 1)
     dividers = np.column_stack(
         (
-            np.zeros((m, 1), dtype=int),
+            np.zeros((m, 1), int),
             np.row_stack(  # type: ignore[call-overload]
                 nchoosek(np.arange(1, k + d), d - 1)
             ),
-            np.full((m, 1), k + d, dtype=int),
+            np.full((m, 1), k + d, int),
         )
     )
     return np.flipud(np.diff(dividers, axis=1) - 1).astype(int)
+
+
+def monomials_basis_function(n_in: int, mindegree: int, maxdegree: int) -> cs.Function:
+    """Creates a function made of monomials as bases.
+
+    Parameters
+    ----------
+    ns : int
+        Number of states.
+    mindegree : int
+        Minimum degree of monomials (included).
+    maxdegree : int
+        Maximum degree of monomials (included).
+
+    Returns
+    -------
+    casadi.Function
+        A casadi function of the form `Phi(s)`, where `s` is the input and `Phi` the
+        basis vector.
+    """
+    s = cs.SX.sym("x", n_in, 1)  # prod is faster with SX, and during runtime
+    y = cs.vertcat(
+        *(
+            prod(s**p)
+            for k in range(mindegree, maxdegree + 1)
+            for p in monomial_powers(n_in, k)
+        )
+    )
+    return cs.Function("Phi", [s], [y], ["s"], ["Phi(s)"])
